/* SPDX-License-Identifier: BSL-1.0 OR BSD-3-Clause */

#ifndef MPT_BASE_INTEGER_HPP
#define MPT_BASE_INTEGER_HPP



#include "mpt/base/namespace.hpp"

#include <cstdint>



namespace mpt {
inline namespace MPT_INLINE_NS {


using int8 = std::int8_t;
using int16 = std::int16_t;
using int32 = std::int32_t;
using int64 = std::int64_t;
using uint8 = std::uint8_t;
using uint16 = std::uint16_t;
using uint32 = std::uint32_t;
using uint64 = std::uint64_t;


} // namespace MPT_INLINE_NS
} // namespace mpt



#endif // MPT_BASE_INTEGER_HPP
