/* SPDX-License-Identifier: BSD-3-Clause */
/* SPDX-FileCopyrightText: OpenMPT Project Developers and Contributors */


#pragma once

#include "openmpt/all/BuildSettings.hpp"

#include "mpt/base/float.hpp"
#include "mpt/base/integer.hpp"



OPENMPT_NAMESPACE_BEGIN



using int8 = mpt::int8;
using int16 = mpt::int16;
using int32 = mpt::int32;
using int64 = mpt::int64;
using uint8 = mpt::uint8;
using uint16 = mpt::uint16;
using uint32 = mpt::uint32;
using uint64 = mpt::uint64;

using nativefloat = mpt::nativefloat;
using somefloat32 = mpt::somefloat32;
using somefloat64 = mpt::somefloat64;
using namespace mpt::float_literals;



OPENMPT_NAMESPACE_END
